@@ -10,7 +10,7 @@ use MIDI::Event;
 use Time::HiRes qw/ time /;
 use Carp qw/ carp croak /;
 
-our $VERSION = '0.06';
+our $VERSION = '0.07';
 
 my $rtmidi_api_names = {
     unspecified => [ "Unknown",            RTMIDI_API_UNSPECIFIED ],
@@ -19,9 +19,9 @@ my $rtmidi_api_names = {
     jack        => [ "Jack",               RTMIDI_API_UNIX_JACK ],
     winmm       => [ "Windows MultiMedia", RTMIDI_API_WINDOWS_MM ],
     dummy       => [ "Dummy",              RTMIDI_API_RTMIDI_DUMMY ],
-    #web         => [ "Web MIDI API",       RTMIDI_API_WEB_MIDI_API ],
-    #winuwp      => [ "Windows UWP",        RTMIDI_API_WINDOWS_UWP ],
-    #amidi       => [ "Android MIDI API",   RTMIDI_API_ANDROID ],
+    web         => [ "Web MIDI API",       RTMIDI_API_WEB_MIDI_API ],
+    winuwp      => [ "Windows UWP",        RTMIDI_API_WINDOWS_UWP ],
+    amidi       => [ "Android MIDI API",   RTMIDI_API_ANDROID ],
 };
 
 my $byte_lookup = {
@@ -200,7 +200,7 @@ sub new {
 }
 
 
-sub ok   { $_[0]->{device}->ok }
+sub ok   { shift->{device}->ok( @_ ) }
 sub msg  { $_[0]->{device}->msg }
 sub data { $_[0]->{device}->data }
 sub ptr  { $_[0]->{device}->ptr }
@@ -208,14 +208,30 @@ sub ptr  { $_[0]->{device}->ptr }
 
 sub open_virtual_port {
     my ( $self, $port_name ) = @_;
+    croak "Virtual ports unsupported on this platform"
+        if $self->get_current_api == RTMIDI_API_WINDOWS_MM;
+    $self->ok(1);
+
     rtmidi_open_virtual_port( $self->{device}, $port_name );
+
+    return 1 if $self->ok;
+
+    croak "Error opening virtual port: " . $self->msg;
 }
 
 
 sub open_port {
     my ( $self, $port_number, $port_name ) = @_;
+    croak "Invalid port number ($port_number)"
+        if ($port_number < 0 || $port_number >= $self->get_port_count);
+    $self->ok(1);
+
     $self->{port_name} = $port_name;
     rtmidi_open_port( $self->{device}, $port_number, $port_name );
+
+    return 1 if $self->ok;
+
+    croak("Error opening port: " . $self->msg);
 }
 
 
@@ -266,7 +282,13 @@ sub get_all_port_names {
 
 sub close_port {
     my ( $self ) = @_;
+    $self->ok(1);
+
     rtmidi_close_port( $self->{device} );
+
+    return 1 if $self->ok;
+
+    croak "Error closing port: " . $self->msg;
 }
 
 
@@ -469,6 +491,7 @@ return_msg:
 
 sub send_message_encoded {
     my ( $self, @event ) = @_;
+    @event = @{ $event[ 0 ] } if ( @event == 1 && ref $event[ 0 ] eq 'ARRAY' );
     if ( $event[0] eq 'control_change' ) {
         my $rpn = $self->get_rpn( $event[1] );
         my $nrpn = $self->get_nrpn( $event[1] );
@@ -525,22 +548,6 @@ sub PANIC {
 }
 
 
-
-*note_off = sub { shift->send_event( note_off => @_ ) };
-*note_on = sub { shift->send_event( note_on => @_ ) };
-*control_change = sub { shift->send_event( control_change => @_ ) };
-*patch_change = sub { shift->send_event( patch_change => @_ ) };
-*key_after_touch = sub { shift->send_event( key_after_touch => @_ ) };
-*channel_after_touch = sub { shift->send_event( channel_after_touch => @_ ) };
-*pitch_wheel_change = sub { shift->send_event( pitch_wheel_change => @_ ) };
-*sysex_f0 = sub { shift->send_event( sysex_f0 => @_ ) };
-*sysex_f7 = sub { shift->send_event( sysex_f7 => @_ ) };
-*sysex = sub { shift->send_event( sysex => @_ ) };
-
-
-*cc = \&control_change;
-
-
 sub get_14bit_mode { $_[0]->{ '14bit_mode' } }
 *get_14bit_callback = \&get_14bit_mode;
 
@@ -633,6 +640,10 @@ sub _create_device {
     $self->{api} //= $api_by_name->[1] if $api_by_name;
     $self->{api} //= $rtmidi_api_names->{ unspecified }->[1];
     $self->{device} = $create_dispatch->{ $fn }->( $self->{api}, $self->{name}, $self->{queue_size_limit} );
+
+    croak "Unable to create MIDI $self->{type} device"
+        unless $self->{device}->ok;
+
     $self->{type} eq 'in' && $self->ignore_types(
         $self->{ignore_sysex},
         $self->{ignore_timing},
@@ -744,6 +755,25 @@ sub disable_nrpn_14bit_mode {
 }
 *disable_nrpn_14bit_callback = \&disable_nrpn_14bit_mode;
 
+
+sub note_off { shift->send_event( note_off => @_ ) };
+sub note_on { shift->send_event( note_on => @_ ) };
+sub control_change { shift->send_event( control_change => @_ ) };
+sub patch_change { shift->send_event( patch_change => @_ ) };
+sub key_after_touch { shift->send_event( key_after_touch => @_ ) };
+sub channel_after_touch { shift->send_event( channel_after_touch => @_ ) };
+sub pitch_wheel_change { shift->send_event( pitch_wheel_change => @_ ) };
+sub sysex_f0 { shift->send_event( sysex_f0 => @_ ) };
+sub sysex_f7 { shift->send_event( sysex_f7 => @_ ) };
+sub sysex { shift->send_event( sysex => @_ ) };
+sub clock { shift->send_event( clock => @_ ) };
+sub start { shift->send_event( start => @_ ) };
+sub stop { shift->send_event( stop => @_ ) };
+sub continue { shift->send_event( continue => @_ ) };
+
+
+*cc = \&control_change;
+
 my $free_dispatch = {
     in  => \&rtmidi_in_free,
     out => \&rtmidi_out_free
@@ -796,7 +826,7 @@ MIDI::RtMidi::FFI::Device - OO interface for MIDI::RtMidi::FFI
 
 =head1 VERSION
 
-version 0.06
+version 0.07
 
 =head1 SYNOPSIS
 
@@ -1097,7 +1127,7 @@ Type 'in' only. Set whether or not to ignore active sensing messages.
 
 Type 'in' only. Gets the next message from the queue, if available.
 
-=head2 get_event
+=head2 get_message_decoded
 
     $device->get_message_decoded();
 
@@ -1200,7 +1230,7 @@ Type 'out' only. Sends a L<MIDI::Event> encoded message to the open port.
     $device->send_message_encoded_cb( @event );
 
 Type 'out' only. A variant of send_message_encoded for use within user-defined
-callbacks, as callbacks are invoked by send_message_encoded.
+callbacks handling 14 bit CC, as callbacks are invoked by send_message_encoded.
 
 =head2 send_event
 
@@ -1230,20 +1260,6 @@ B<Warning:> This method has the potential to flood buffers!
 It should be a recourse of last resort - consider L</panic>,
 it'll probably work.
 
-=head2 note_off, note_on, control_change, patch_change, key_after_touch, channel_after_touch, pitch_wheel_change, sysex_f0, sysex_f7, sysex
-
-Wrapper methods for L</send_message_encoded>, e.g.
-
-    $device->note_on( 0x00, 0x40, 0x5a );
-
-is equivalent to:
-
-    $device->send_message_encoded( note_on => 0x00, 0x40, 0x5a );
-
-=head2 cc
-
-An alias for control_change.
-
 =head2 get_14bit_mode
 
     $device->get_14bit_mode;
@@ -1265,7 +1281,7 @@ or decoding mode - it probably shouldn't be used mid performance or
 playback unless you seek odd side effects.
 
 If a RPN or NRPN is active, this 14 bit mode will not have an effect on CC6.
-See </set_set_rpn_14bit_mode> and </set_nrpn_14bit_mode>
+See L</set_set_rpn_14bit_mode> and L</set_nrpn_14bit_mode>
 
 =head2 disable_14bit_mode
 
@@ -1350,13 +1366,13 @@ Close any open RPN on the given channel.
 
 Close any open NRPN on the given channel.
 
-=head1 get_rpn
+=head2 get_rpn
 
     $device->get_nrpn( $channel );
 
 Get the currently open RPN for the given channel.
 
-=head1 get_nrpn
+=head2 get_nrpn
 
     $device->get_rpn( $channel );
 
@@ -1446,6 +1462,20 @@ Disables the RPN 14 bit mode. See L</14-bit Control Change Modes>.
 
 Disables the NRPN 14 bit mode. See L</14-bit Control Change Modes>.
 
+=head2 note_off, note_on, control_change, patch_change, key_after_touch, channel_after_touch, pitch_wheel_change, sysex_f0, sysex_f7, sysex, clock, start, stop, continue
+
+Wrapper methods for L</send_message_encoded>, e.g.
+
+    $device->note_on( 0x00, 0x40, 0x5a );
+
+is equivalent to:
+
+    $device->send_message_encoded( note_on => 0x00, 0x40, 0x5a );
+
+=head2 cc
+
+An alias for control_change.
+
 =head1 14 bit Control Change Modes
 
 14 bit Control Change messages are achieved by sending a pair of 7-bit
@@ -1515,7 +1545,7 @@ again."> - again, this is fine - it fits in with expectations so far.
 
 I<"When an MSB is received, the receiver should set its concept of the LSB
 to zero">. This, to me, is ambiguous. Should our CC now be set to
-MSB << 7 + 0? Or is it an instruction to forget any existing LSB value and
+C<( $msb << 7 ) + 0>? Or is it an instruction to forget any existing LSB value and
 await the transmission of a fresh one before constructing a CC value?
 
 With the former approach you could imagine a descending control passing a
@@ -1533,14 +1563,14 @@ MSB threshold (a large jump in LSB).
 
 Some implementations send LSB first, MSB second. If a LSB/MSB pair is sent
 each time, this is easily handled. If a pair is sent, then fine control
-is sent via LSB we have a problem. When we cross a MSB threshold,
+is sent subsequently via LSB we have a problem. When we cross a MSB threshold,
 we need to wait for the new MSB value before we can construct the complete
 CC value. This means we need to somehow know when to stop performing fine
 control with new LSB values, and await a new MSB value - we are back to
 heuristic detection, looking for LSB jumps.
 
 All to say, there are some ambiguities in how this is handled, and there
-are endless variations between devices.
+are endless variations between different devices and implementations.
 
 The second problem is needing to write explicit 14 bit message handling in
 each project individually. This module intends to obviate some of this by
@@ -1636,7 +1666,7 @@ MIDI standard:
 Callbacks should not call the send_message_encoded, send_event, control_change
 or cc methods as these may invoke further 14 bit message handling, potentially
 causing an
-infinite loop. The </send_message_encoded_cb> method exists for sending
+infinite loop. The L</send_message_encoded_cb> method exists for sending
 messages within 14 bit CC callbacks.
 
 =head2 For Input (Decoding)
@@ -1656,7 +1686,7 @@ reset to zero and a new value is returned.
 
 =head3 await (recommended)
 
-This is the same as 'midi' mode, but it aleays awaits a LSB message before
+This is the same as 'midi' mode, but it always awaits a LSB message before
 returning a value.
 
 This is likely the most compatible and reliable mode for decoding.
@@ -1799,14 +1829,29 @@ Channel must be specified in any message related to performance, such as
 
 =head2 Messages and Events
 
-A MIDI message is (usually) a short series of bytes sent on a port, instructing
+A MIDI message is a (usually) short series of bytes sent on a port, instructing
 an instrument on how to behave - which notes to play, when, how loudly, with which
 timbral variations & expression, and so on. They may also contain configuration
 info or some other sort of instruction.
 
 In this module "events" usually refer to incoming message bytes decoded into a
-descriptive sequence of values, or a mechanism for turning these values into
-message bytes for ouput.
+descriptive sequence of values, or a mechanism for turning these descriptive
+sequences into message bytes for ouput.
+
+=head2 General MIDI and Soundfonts
+
+General MIDI is a specification which standardises a single set of musical
+instruments, accessed via the "patch change" command. Any of 128 instruments
+may be assigned to any of 16 channels, with the exception of channel 10
+(0x09) which is reserved for percussion.
+
+Soundfonts are banks of sampled instruments which may be loaded by a General
+MIDI compatible softsynth. These can be quite large and complex, though they
+usually tend to be small and cheesy. If you remember 90s video game
+music or web pages playing .mid files, you're on the right track.
+
+Some implementations also support DLS files, which are similar to soundfonts,
+though unlike soundfonts the specification is freely available.
 
 =head1 Virtual Devices and Windows
 
@@ -1880,7 +1925,7 @@ distro. To run FluidSynth you'll need a SF2 or SF3 soundfont file. See
 L<Getting started with fluidsynth|https://github.com/FluidSynth/fluidsynth/wiki/GettingStarted>
 and
 L<Example Command Lines to start fluidsynth|https://github.com/FluidSynth/fluidsynth/wiki/ExampleCommandLines>.
-<FluidR3_GM.sf2 Professional|https://musical-artifacts.com/artifacts/738>
+L<FluidR3_GM.sf2 Professional|https://musical-artifacts.com/artifacts/738>
 is a high quality sound font with a complete set of General MIDI instruments.
 
 A typical FluidSynth invocation on Linux might be:
@@ -1890,10 +1935,10 @@ A typical FluidSynth invocation on Linux might be:
 =head1 General MIDI on MacOS
 
 An Audio Unit named DLSMusicDevice is available for use within GarageBand,
-Logic, and other Digital Audio Workstation (DAW) software on Mac OS X.
+Logic, and other Digital Audio Workstation (DAW) software on MacOS.
 
 If you wish to use banks other than the default QuickTime set, place
-them in ~/Library/Audio/Sounds/Banks/. You may now create a new track
+them in C<~/Library/Audio/Sounds/Banks/>. You may now create a new track
 within GarageBand or Logic with the DLSMusicDevice instrument, and
 select your Sound Bank within the settings for this instrument.
 
@@ -1910,16 +1955,20 @@ your DAW and be ready to send performance info to DLSMusicDevice:
 The 'MUS 214: MIDI Composition' channel on YouTube has a
 L<Video on setting up DLSMusicDevice in Logic|https://youtu.be/YIb-H10yzyI>.
 
-A potential alternative option is FluidSynth. This has more limited support
-for DLS patches but should load SF2/3 banks just fine. See
-L</General MIDI on Linux> for links to get started using FluidSynth.
-
-A typical FluidSynth invocation on MacOS might be:
+A potential alternative option is FluidSynth. This has more limited support for
+DLS banks but should load SF2/3 banks just fine. See L</General MIDI on Linux>
+for links to get started using FluidSynth. A typical FluidSynth invocation on
+MacOS might be:
 
     % fluidsynth -a coreaudio -m coremidi your_soundfont.sf2
 
 =head1 KNOWN ISSUES
 
+The callback interface does not currently work on threaded perls. Most, if not
+all, perls currently built for Windows are threaded. I have been working around
+this with a non-threaded Perl built within the cygwin environment with
+perlbrew.
+
 Use of L<MIDI::Event> is a bit of a hack for convenience, exploiting the
 similarity of realtime MIDI messages and MIDI song file messages. It may break
 in unexpected ways if used for large SysEx messages or other "non-music"
@@ -1948,6 +1997,8 @@ L<RtMidi|https://www.music.mcgill.ca/~gary/rtmidi/>
 
 L<MIDI CC & NRPN database|https://midi.guide/>
 
+L<Sound on Sound's MIDI Basics series|https://www.soundonsound.com/series/midi-basics>
+
 L<Phil Rees Music Tech page on NRPN/RPN|http://www.philrees.co.uk/nrpnq.htm>
 
 L<MIDI::RtMidi::FFI>
@@ -1980,7 +2031,7 @@ John Barrett <john@jbrt.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2024 by John Barrett.
+This software is copyright (c) 2025 by John Barrett.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
